# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2023 - Aleks Kissinger, John van de Wetering,
#                      and Arianne Meijer-van de Griend

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fractions import Fraction
from typing import Callable, Dict, List, Set, Tuple, Union, Optional, Any
from enum import Enum

from ..circuit import Circuit, ZPhase, HAD, XPhase, CNOT
from ..graph.graph import Graph
from ..linalg import Mat2, MatLike
from ..routing.parity_maps import CNOT_tracker, Parity
from ..routing.cnot_mapper import sequential_gauss, ElimMode, gauss
from ..routing.steiner import steiner_reduce_column
from ..routing.architecture import create_architecture, FULLY_CONNECTED, Architecture
from ..utils import make_into_list, maxelements


class RoutingMethod(Enum):
    """
    Phase polynomial routing method to use in :func:`.route_phase_poly`.
    """

    MATROID = "matroid"
    """Routing method based on matroid partitioning. Commonly slower than
    :attr:`.RoutingMethod.GRAY` and :attr:`.RoutingMethod.MEIJER`."""

    GRAY = "GraySynth"
    """
    Routing method based on Gray synthesis (see arxiv.org/abs/1712.01859 ).
    """

    MEIJER = "meijer"
    """
    Routing method by Meijer and Duncan (see arxiv.org/abs/2004.06052 ).
    """

    GRAY_MEIJER = "GraySynth+Meijer"
    """
    Combination of :attr:`.RoutingMethod.GRAY` and :attr:`.RoutingMethod.MEIJER`, keeps the best result of both.
    """

    def __str__(self):
        return f"{self.value}"


class RootHeuristic(Enum):
    """
    Heuristics for choosing the root of a Steiner tree during phase polynomial routing.
    """

    RANDOM = "gauss"
    """Randomly choose a root."""

    EXHAUSTIVE = "exhaustive"
    """Try all possible roots and choose the one with the lowest cost."""

    ARITY = "arity"
    """Choose the root randomly between the nodes with highest arity."""

    RECURSIVE = "recursive"
    """Use an already-chosen root in a recursive call."""

    def __str__(self):
        """
        Converts RootHeuristic into a string
        """
        return f"{self.value}"

    def to_function(
        self,
    ) -> Callable[[Architecture, Mat2, List[int], List[int], int, int, Any], List[int]]:
        """
        Looks at the RootHeuristic type and returns the relevant root heuristic
        """
        # ignore types due to https://github.com/python/mypy/issues/5876
        if self == RootHeuristic.RANDOM:
            return random_root_heuristic  # type: ignore
        elif self == RootHeuristic.EXHAUSTIVE:
            return exhaustive_root_heuristic  # type: ignore
        elif self == RootHeuristic.ARITY:
            return arity_root_heuristic  # type: ignore
        elif self == RootHeuristic.RECURSIVE:
            return rec_root_heuristic  # type: ignore
        else:
            raise KeyError(f"The root heuristic '{self}' is not implemented")


class SplitHeuristic(Enum):
    """
    Heuristics for choosing nodes to split a circuit during phase polynomial routing.
    """

    RANDOM = "random"
    """Randomly pick a candidate."""

    ARITY = "arity"
    """Split the circuit on the nodes with highest arity."""

    COUNT = "count"
    """Split the circuit on all the candidate nodes."""

    def __str__(self):
        """
        Converts SplitHeuristic into a string
        """
        return f"{self.value}"

    def to_function(
        self,
    ) -> Callable[[Architecture, Mat2, List[int], List[int], Any], List[int]]:
        """
        Looks at the SplitHeuristic type and returns the relevant split heuristic
        """
        if self == SplitHeuristic.RANDOM:
            return random_split_heuristic  # type: ignore
        elif self == SplitHeuristic.ARITY:
            return arity_split_heuristic  # type: ignore
        elif self == SplitHeuristic.COUNT:
            return count_split_heuristic  # type: ignore
        else:
            raise KeyError(f"The split heuristic '{self}' is not implemented")


def route_phase_poly(
    circuit: Union[Circuit, "PhasePoly"],
    architecture: Architecture,
    method: RoutingMethod = RoutingMethod.GRAY_MEIJER,
    mode: ElimMode = ElimMode.STEINER_MODE,
    root_heuristic: RootHeuristic = RootHeuristic.RECURSIVE,
    split_heuristic: SplitHeuristic = SplitHeuristic.COUNT,
    **kwargs,
) -> Circuit:
    """
    Compile a circuit to an architecture with restricted connectivity.

    :param circuit: The circuit to compile.
    :param architecture: The target architecture.
    :param method: The routing method to use.
    :param mode: The elimination mode to use during the CNOT mapping step.
    :param split_heuristic: The heuristic to use for splitting the circuit into subcircuits.
    :param root_heuristic: The heuristic to use for finding the root of the circuit.
    :return: The compiled circuit.
    """
    if isinstance(circuit, Circuit):
        phase_poly = PhasePoly.fromCircuit(circuit)
    else:
        phase_poly = circuit
    new_circuit = None
    if method == RoutingMethod.MATROID:
        new_circuit = phase_poly.matroid_synth(mode, architecture, **kwargs)[0]
    elif method == RoutingMethod.MEIJER:
        new_circuit = phase_poly.Ariannes_synth(mode, architecture, **kwargs)[0]
    elif method == RoutingMethod.GRAY_MEIJER:
        circuit1 = phase_poly.Ariannes_synth(mode, architecture, **kwargs)[0]
        circuit2 = phase_poly.rec_gray_synth(
            mode,
            architecture,
            split_heuristic=split_heuristic,
            root_heuristic=root_heuristic,
            **kwargs,
        )[0]
        if circuit1.twoqubitcount() < circuit2.twoqubitcount():
            new_circuit = circuit1
        else:
            new_circuit = circuit2
    elif method == RoutingMethod.GRAY:
        new_circuit = phase_poly.rec_gray_synth(
            mode,
            architecture,
            split_heuristic=split_heuristic,
            root_heuristic=root_heuristic,
            **kwargs,
        )[0]
    else:
        raise KeyError(f"The routing method '{method}' is not supported")
    return new_circuit


def random_root_heuristic(
    architecture, matrix, cols_to_use, qubits, column, phase_qubit, **kwargs
):
    """
    Randomly chooses a root from the given qubits and creates a steiner reduction for a specific column in the matrix

    :param architecture: The quantum architecture to be used
    :param matrix: The matrix to be used
    :param cols_to_use: (not directly used)
    :param qubits: The list of qubits to randomly select a root from
    :param column: Column to be reduced
    :param phase_qubit: (not directly used)
    :param **kwargs: Additional arguments passed
    :return: The sequence of row operations applied
    """
    root = np.random.choice(qubits)
    return list(
        steiner_reduce_column(
            architecture,
            [row[column] for row in matrix.data],
            root,
            qubits,
            [i for i in range(architecture.n_qubits)],
            [],
            upper=True,
        )
    )


def exhaustive_root_heuristic(
    architecture, matrix, cols_to_use, qubits, column, phase_qubit, **kwargs
):
    """
    Exhastively applies the Steiner reduction with all the given qubits as roots for a specific column in the matrix and returns the shortest 
    reduction sequence.

    :param architecture: The quantum architecture to be used
    :param matrix: The matrix to be used
    :param cols_to_use: (not directly used)
    :param qubits: The list of qubits to be used as roots
    :param column: Column to be reduced
    :param phase_qubit: (not directly used)
    :param **kwargs: Additional arguments passed
    :return: The shortest sequence of row operations applied
    """
    cnots = None
    for root in qubits:
        # build a steiner tree
        # Steiner extract with that qubit as root
        possible_cnots = list(
            steiner_reduce_column(
                architecture,
                [row[column] for row in matrix.data],
                root,
                qubits,
                [i for i in range(architecture.n_qubits)],
                [],
                upper=True,
            )
        )
        if cnots is None or len(possible_cnots) < len(cnots):
            cnots = possible_cnots
    return cnots


def arity_root_heuristic(
    architecture, matrix, cols_to_use, qubits, column, phase_qubit, **kwargs
):
    """
    Filters qubits by highest arity and randomly selects one as a root for a specific column in the matrix and returns the 
    reduction sequence.

    :param architecture: The quantum architecture to be used
    :param matrix: The matrix to be used
    :param cols_to_use: (not directly used)
    :param qubits: The list of qubits to be used as roots
    :param column: Column to be reduced
    :param phase_qubit: (not directly used)
    :param **kwargs: Additional arguments passed
    :return: The sequence of row operations applied
    """
    iterator = (
        (qubit, arity) for qubit, arity in architecture.arities() if qubit in qubits
    )
    q, a = next(iterator)
    best_qubits = []
    best_arity = a
    while a == best_arity:
        best_qubits.append(q)
        q, a = next(iterator, (None, best_arity - 1))
    root = np.random.choice(best_qubits)
    return list(
        steiner_reduce_column(
            architecture,
            [row[column] for row in matrix.data],
            root,
            qubits,
            [i for i in range(architecture.n_qubits)],
            [],
            upper=True,
        )
    )


def rec_root_heuristic(
    architecture, matrix, cols_to_use, qubits, column, phase_qubit, **kwargs
):
    """
    Uses the phase_qubit as a root for a specific column in the matrix and returns the reduction sequence.

    :param architecture: The quantum architecture to be used
    :param matrix: The matrix to be used
    :param cols_to_use: (not directly used)
    :param qubits: The list of qubits to be used as roots
    :param column: Column to be reduced
    :param phase_qubit: The qubit to be used as the root
    :param **kwargs: Additional arguments passed
    :return: The sequence of row operations applied
    """
    root = phase_qubit
    return list(
        steiner_reduce_column(
            architecture,
            [row[column] for row in matrix.data],
            root,
            qubits,
            [i for i in range(architecture.n_qubits)],
            [],
            upper=True,
        )
    )


def calculate_reward(architecture, matrix, cols_to_use, root):
    """
    Runs Steiner reduction over multiple columns and sums up the total number of CNOT operations, outputting a reward score

    :param architecture: The quantum architecture to be used
    :param matrix: The matrix to be used
    :param cols_to_use: A list of columns to be used in calculating the reward
    :param root: The qubit to be used as the root

    :return: The reward score
    """
    n_qubits = architecture.n_qubits
    all_cnots = 0
    for column in cols_to_use:
        qubits = [
            i
            for i in range(n_qubits)
            if sum([matrix.data[i][j] for j in cols_to_use]) == len(cols_to_use)
        ]
        cnots = list(
            steiner_reduce_column(
                architecture,
                [row[column] for row in matrix.data],
                root,
                qubits,
                [i for i in range(n_qubits)],
                [],
                upper=True,
            )
        )
        all_cnots += len(cnots)
    return 2 * n_qubits - all_cnots / (len(cols_to_use) + 1)


def random_split_heuristic(
    architecture: Architecture,
    matrix: Mat2,
    cols_to_use: List[int],
    qubits: List[int],
    **kwargs,
):
    """
    Returns a random split of qubits

    :param architecture: (Not directly used)
    :param matrix: (Not directly used)
    :param cols_to_use: (Not directly used)
    :param qubits: The list of qubits to be split
    :param **kwargs: Additional arguments passed

    :return: A random split from the given qubits
    """
    return [np.random.choice(qubits)]


def arity_split_heuristic(architecture, matrix, cols_to_use, qubits, **kwargs):
    """
    Returns a split of qubits with the highest arity score

    :param architecture: The quantum architecture to be used
    :param matrix: (Not directly used)
    :param cols_to_use: (Not directly used)
    :param qubits: The list of qubits to be split
    :param **kwargs: Additional arguments passed

    :return: A split from the given qubits with the highest arity score
    """
    iterator = (
        (qubit, arity) for qubit, arity in architecture.arities() if qubit in qubits
    )
    q, a = next(iterator)
    best_qubits = []
    best_arity = a
    while a == best_arity:
        best_qubits.append(q)
        q, a = next(iterator, (None, best_arity - 1))
    return best_qubits


def count_split_heuristic(architecture, matrix, cols_to_use, qubits, **kwargs):
    """
    Returns a split of qubits with the highest arity score

    :param architecture: (Not directly used)
    :param matrix: (Not directly used)
    :param cols_to_use: (Not directly used)
    :param qubits: The list of qubits to be split
    :param **kwargs: Additional arguments passed

    :return: The qubits passed in
    """
    return qubits


class PhasePoly:
    """
    A class representing a phase polynomial.
    """

    def __init__(
        self,
        zphase_dict: Dict[Parity, Fraction],
        out_parities: List[Parity],
    ):
        """
        Creates and returns a phase polynomial object.

        :param zphase_dict: A dictionary of parity objects and phase factors
        :param out_parities: A list of final parity objects
        """
        self.zphases = zphase_dict
        self.out_par = out_parities
        self.n_qubits = out_parities[0].n_qubits()
        self.all_parities = list(zphase_dict.keys())

    @staticmethod
    def fromCircuit(
        circuit,
        initial_qubit_placement=None,
        final_qubit_placement=None,
    ):
        """
        Takes a circuit and tracks how qubit parities evolve as gates are applied, 
        accumulates Z-phase rotations then builds a compact PhasePoly representing the circuit's phase structure

        :param circuit: The circuit to be used
        :param initial_qubit_placement: Optional list specifying initial qubit mapping, default None
        :param final_qubit_placement: Optional list specifying final qubit mapping, default None
        :return: A PhasePoly object representing the circuit's accumulated Z-phase structure
        """
        zphases = {}
        current_parities = mat22partition(Mat2.id(circuit.qubits))
        if initial_qubit_placement is not None:
            current_parities = [
                Parity(row[i] for i in initial_qubit_placement)
                for row in current_parities
            ]
        for gate in circuit.gates:
            parity = current_parities[gate.target]
            if gate.name in ["CNOT", "CX"]:
                # Update current_parities
                control = current_parities[gate.control]
                current_parities[gate.target] = Parity(
                    (int(i) + int(j)) % 2 for i, j in zip(control, parity)
                )
            elif isinstance(gate, ZPhase):
                # Add the T rotation to the phases
                if parity in zphases:
                    zphases[parity] += gate.phase
                else:
                    zphases[parity] = gate.phase
            elif isinstance(gate, XPhase):
                raise Exception("XPhase gate not yet supported!")
            else:
                raise Exception("Gate not supported!", gate.name)

        def clamp(phase):
            """
            Normalize phase into range [-1, 1]

            :param phase: The initial phase
            :return: The new phase
            """
            new_phase = phase % 2
            if new_phase > 1:
                return new_phase - 2
            return new_phase

        zphases = {par: clamp(r) for par, r in zphases.items() if clamp(r) != 0}
        if final_qubit_placement is not None:
            current_parities = [current_parities[i] for i in final_qubit_placement]

        return PhasePoly(zphases, current_parities)

    def partition(self, skip_output_parities=True, optimize_parity_order=False):
        """
        Uses matroid partitioning algorithm to split the parities into independent partiotions

        :param skip_output_parities: If True, ignore the out_par parities, default True
        :param optimize_parity_order: If True, reorder parities for more optimal placement, default False
        :return: List of ordered parity partitions
        """
        # Matroid partitioning based on wikipedia: https://en.wikipedia.org/wiki/Matroid_partitioning
        def add_edge(graph, vs_dict, node1, node2):
            v1 = [k for k, v in vs_dict.items() if v == node1][0]
            v2 = [k for k, v in vs_dict.items() if v == node2][0]
            graph.nedges += 1
            graph.graph[v1][v2] = 1

        partitions = []
        parities_to_partition = (
            self.all_parities
            if not skip_output_parities
            else [p for p in self.all_parities if p not in self.out_par]
        )
        for parity in parities_to_partition:
            graph = Graph()
            # Add current parity to the graph
            # Add each partition to the graph
            # Add each parity in the partition to the graph
            vs = graph.add_vertices(
                len(partitions) + sum([len(p) for p in partitions]) + 1
            )
            vs_dict = {
                i: node
                for i, node in zip(
                    vs, partitions + [e for p in partitions for e in p] + [parity]
                )
            }
            # Check which parities can be added to another partition
            for partition in partitions:
                # Check if the new parity can be added to the partition
                if self._independent(list(partition) + [parity]):
                    add_edge(graph, vs_dict, parity, partition)
                for p in partition:
                    # Check if p can be replaced with the new parity in the partition
                    for partition2 in partitions:
                        if partition2 != partition:
                            for p2 in partition2:
                                # Check if p2 can be replaced with p in their partitions
                                new_partition = [e for e in partition if e != p] + [p2]
                                if self._independent(new_partition):
                                    add_edge(graph, vs_dict, p2, p)
            # Find a path from the parity to a partition
            path = self._dfs([(parity, [parity])], graph, vs_dict, partitions)
            if path != []:
                # Apply those changes if such a path exists
                # Remember which partition to add the final element to
                p_idx = partitions.index(
                    path[-1]
                )  # Last element in the path is always a partition
                for p1, p2 in zip(path[:-1], path[1:]):
                    # Replace p2 with p1 in p1's partition
                    for partition in partitions:
                        if p1 in partition:
                            partition.pop(p1)
                            partition.append(p2)
                partitions[p_idx].append(path[-2])
            else:
                # Make a new partition if no such path exists.
                partitions.append([parity])
        ordered_partitions = []
        for partition in partitions:
            # Add identity parities to the partition if the set is not full yet.
            matrix = partition2mat2(partition)
            inv = inverse_hack(matrix)
            if inv is not None:
                matrix, _ = inv
            if optimize_parity_order:
                # Find a more optimal ordering of the parities
                parity_placement = self._place_parities(matrix)
            else:
                parity_placement = np.arange(self.n_qubits)
            partition = mat22partition(matrix)
            new_partition = [partition[i] for i in parity_placement]
            ordered_partitions.append(new_partition)
        return ordered_partitions

    def _order_partitions(self, partitions):
        """
        Reorders partitions based on a cost metric

        :param partitions: List of partitions, where each partition is a list of Parity objects
        :return: Partitions ordered by cost metric
        """
        n = len(partitions)
        numbered = {
            i: partition2mat2(partition) for i, partition in enumerate(partitions)
        }

        def cost_func(p1, p2):
            """
            Calculates the cost to go from one partition to another

            :param p1: First partition
            :param p2: Second partition
            :return: Cost to go from p1 to p2
            """
            inv = inverse_hack(p1)
            if inv is not None:
                _, inv = inv
            else:
                inv = Mat2.id(p2.rows())
            c = CNOT_tracker(len(partitions[0]))
            (p2 * inv).gauss(full_reduce=True, y=c)
            return len(c.gates)

        path_costs = {
            (i, j): cost_func(p1, p2)
            for i, p1 in numbered.items()
            for j, p2 in numbered.items()
            if i != j
        }
        # start at the back
        new_partitions = [partitions[-1]]
        visited = []
        current = n - 1
        while len(visited) < n - 1:
            # Pick the partition that was not yet visited whose
            choice = min(
                [i for i in range(n) if i not in visited + [current]],
                key=lambda i: path_costs[(i, current)],
            )
            new_partitions = [mat22partition(numbered[choice])] + new_partitions
            visited += [current]
            current = choice
        return new_partitions

    def _place_parities(self, parities: Union[Mat2, MatLike]) -> List[int]:
        """
        Assigns parities to qubits

        :param parities: A list representing each parity
        :return: List of indices showing which parity should be assigned to which qubit index
        """
        if isinstance(parities, Mat2):
            parities = parities.data
        permutation = [0] * self.n_qubits
        skipped_parities = []
        # Greedily map identity rows
        for i, parity in enumerate(parities):
            if parity.count(1) == 1:
                permutation[parity.index(1)] = i
            else:
                skipped_parities.append((i, parity))
        skipped_idxs = []
        deliberating = []
        for i in range(self.n_qubits):
            if permutation[i] is None:
                # Find the best parity in skipped_parities to place there
                pivoted_parities = [(j, p) for j, p in skipped_parities if p[i] == 1]
                if pivoted_parities:
                    # The parity with the least 1s has the most priority because it can probably not be placed elsewhere
                    pivoted_parities = sorted(
                        pivoted_parities, key=lambda parity: parity[1].count(1)
                    )
                    if len(pivoted_parities) == 1:
                        chosen = pivoted_parities[0]
                        permutation[i] = chosen[0]
                        skipped_parities.remove(chosen)
                    else:
                        # If there are multiple options, try them later
                        deliberating.append((i, pivoted_parities))
                else:
                    skipped_idxs.append(i)
        for i, pivoted_parities in deliberating:
            pivoted_parities = [p for p in pivoted_parities if p in skipped_parities]
            if pivoted_parities:
                chosen = pivoted_parities[0]
                permutation[i] = chosen[0]
                skipped_parities.remove(chosen)
            else:
                skipped_idxs.append(i)
        for i in skipped_idxs:
            permutation[i] = skipped_parities.pop()[0]
        return permutation

    def _dfs(self, nodes, graph, inv_vs_dict, partitions):
        """
        A depth first search to find the path

        :param nodes: A list of (current_node, current_path) tuples
        :param graph: The graph showcasing the adjacency connections
        :param inv_vs_dict: A reverse lookup
        :param partitions: List of node objects considered as target nodes
        :return: The path
        """
        # recursive dfs to find the shortest path
        if nodes == []:
            return []
        new_nodes = []
        for (node, path) in nodes:
            # For each node2 connected to node
            v = [k for k, v in inv_vs_dict.items() if v == node][0]
            for node2 in graph.graph[v].keys():
                # If it's a partition, we found a path
                next_node = inv_vs_dict[node2]
                if next_node not in path:  # Avoid loops!
                    new_path = path + [next_node]
                    if next_node in partitions:
                        return new_path
                    else:
                        new_nodes.append((next_node, new_path))
        return self._dfs(new_nodes, graph, inv_vs_dict, partitions)

    @staticmethod
    def _independent(partition: List[Parity]) -> bool:
        """
        Checks if a partition is independent

        :param partition: The list of parities to check
        :return: True if partition is independent 
        """
        return inverse_hack(partition2mat2(partition)) is not None

    def matroid_synth(
        self,
        mode: ElimMode,
        architecture: Architecture,
        optimize_parity_order: bool = False,
        optimize_partition_order: bool = True,
        iterative_placement: bool = False,
        parity_permutation: bool = True,
        iterative_initial_placement: bool = False,
        **kwargs,
    ):
        """
        Using Matroid synthesis method, partitions parities, builds transformation matrices, runs sequential Gaussian elimination,
        builds a final circuit, tracks and updates the current parity states.

        :param mode: Elimination mode, controls Gaussian elimination strategy
        :param architecture: The quantum architecture to be used
        :param optimize_parity_order: Reorder parities within each partition, default False
        :param optimize_partition_order: Reorder the whole partition sequence, default True
        :param iterative_placement: Place qubits iteratively during synthesis, default False
        :param parity_permutation: Allow permuting parities for better optimization, default True
        :param iterative_initial_placement: Optimize initial qubit placement iteratively, default False
        :param **kwargs: Additional arguments passed
        :return: The full Circuit, the input permutation, the output permutation
        """
        kwargs["full_reduce"] = True
        n_qubits = (
            architecture.n_qubits if architecture is not None else len(self.out_par)
        )
        # Partition and order the parities
        partitions = self.partition(optimize_parity_order=optimize_parity_order) + [
            self.out_par
        ]
        if optimize_partition_order:
            partitions = self._order_partitions(partitions)
        # Make the parity sets into matrices
        matrices = [partition2mat2(partition) for partition in partitions]
        # The matrices to be computed need to first undo the previous parities and then obtain the new parities
        other_matrices = []
        prev_matrix = Mat2.id(n_qubits)
        for m in matrices:
            inv = inverse_hack(m)
            if inv is not None:
                new_matrix, inverse = inv
                other_matrices.append(new_matrix * prev_matrix)
                prev_matrix = inverse
            else:
                new_matrix = m
                other_matrices.append(new_matrix * prev_matrix)
                prev_matrix = Mat2.id(n_qubits)
        CNOT_circuits, perms, _ = sequential_gauss(
            [m.copy() for m in other_matrices],
            mode=mode,
            architecture=architecture,
            **kwargs,
        )
        zphases = list(self.zphases.keys())
        circuit = Circuit(n_qubits)
        # Keep track of the parities
        current_parities = Mat2(
            [
                [
                    Mat2.id(n_qubits).data[i][perms[0].index(j)]
                    for j in range(n_qubits)
                ]
                for i in range(n_qubits)
            ]
        )
        for c in CNOT_circuits:
            # Obtain the specified parity
            for gate in c.gates:
                # CNOTs have been mapped already, do not need to be adjusted!
                circuit.add_gate(gate)
                try:
                    current_parities.row_add(gate.control, gate.target)  # type: ignore
                except AttributeError:
                    pass
            # Place the rotations at each parity
            for target, p in enumerate(current_parities.data):
                parity = Parity(p)
                # Apply the phases at current parity if needed.
                if parity in zphases:
                    phase = self.zphases[parity]
                    gate = ZPhase(target=target, phase=phase)
                    zphases.remove(parity)
                    circuit.add_gate(gate)
        # Return the circuit
        return circuit, perms[0], perms[-1]

    def gray_synth(
        self, mode: ElimMode, architecture: Optional[Architecture], **kwargs
    ):
        """
        Using gray synthesis method, builds a quantum circuit by reducing a parity matrix

        :param mode: Elimination mode, controls Gaussian elimination strategy
        :param architecture: An optional of the quantum architecture to be used
        :param **kwargs: Additional arguments passed
        :return: Constructed circuit, input and output qubit orderings
        """
        kwargs["full_reduce"] = True
        arch: Architecture
        if architecture is None or mode == ElimMode.GAUSS_MODE:
            arch = create_architecture(FULLY_CONNECTED, n_qubits=len(self.out_par[0]))
        else:
            arch = architecture
            mode = (
                ElimMode.GAUSS_MODE
                if mode == ElimMode.GAUSS_MODE
                else ElimMode.STEINER_MODE
            )
        # Obtain the parities
        parities_to_reach = self.all_parities
        # TODO - Pick a good order
        # Make a matrix from the parities
        matrix = Mat2(
            [
                [1 if parity[i] else 0 for parity in parities_to_reach]
                for i in range(arch.n_qubits)
            ]
        )
        circuit = CNOT_tracker(arch.n_qubits)
        # For each column in the matrix Or while the matrix has columns
        cols_to_skip = []
        n_phases_placed = 0
        for col in [c for c in range(matrix.cols()) if c not in cols_to_skip]:
            if sum([row[col] for row in matrix.data]) == 1:
                # Add phase gates where needed
                qubit = [row[col] for row in matrix.data].index(1)
                circuit.add_gate(ZPhase(qubit, self.zphases[parities_to_reach[col]]))
                n_phases_placed += 1
                # Remove columns from the matrix if the corresponding parity was obtained
                cols_to_skip.append(col)
        for c in range(matrix.cols()):
            if c not in cols_to_skip:
                column = [int(row[c]) for row in matrix.data]
                # TODO - Pick a good qubit where the phase should be placed
                root = column.index(
                    1
                )  # Place the parity on the first qubit that has a 1
                # Steiner extract with that qubit as root
                nodes = [i for i in range(arch.n_qubits) if column[i] == 1]
                cnots = list(
                    steiner_reduce_column(
                        arch,
                        column,
                        root,
                        nodes,
                        [i for i in range(arch.n_qubits)],
                        [],
                        upper=True,
                    )
                )
                # For each returned CNOT:
                for target, control in cnots:
                    # Place the CNOT on the circuit
                    circuit.add_gate(CNOT(control, target))
                    # Adjust the matrix accordingly - reversed elementary row operations
                    matrix.row_add(target, control)
                    # Keep track of the parities in the circuit - normal elementary row operations
                    for col in [
                        c for c in range(matrix.cols()) if c not in cols_to_skip
                    ]:
                        if sum([row[col] for row in matrix.data]) == 1:
                            # Add phase gates where needed
                            qubit = [row[col] for row in matrix.data].index(1)
                            circuit.add_gate(
                                ZPhase(qubit, self.zphases[parities_to_reach[col]])
                            )
                            n_phases_placed += 1
                            # Remove columns from the matrix if the corresponding parity was obtained
                            cols_to_skip.append(col)
        # Calculate the final parity that needs to be added from the circuit and self.out_par
        self._obtain_final_parities(circuit, arch, mode, **kwargs)
        # Return the circuit
        return (
            circuit,
            [i for i in range(arch.n_qubits)],
            [i for i in range(arch.n_qubits)],
        )

    def rec_gray_synth(
        self,
        mode,
        architecture,
        root_heuristic: RootHeuristic = RootHeuristic.RECURSIVE,
        split_heuristic: SplitHeuristic = SplitHeuristic.COUNT,
        full=True,
        phase_qubit=None,
        **kwargs,
    ):
        """
        Builds quantum circuit by applying a recursive Gray synthesis strategy

        :param mode: Elimination mode, controls Gaussian elimination strategy
        :param architecture: An optional of the quantum architecture to be used
        :param root_heuristic: Controls how to pick the root qubit when extracting Steiner trees
        :param split_heuristic: Controls how to split the matrix into subproblems
        :param full: Whether to apply final parity fixes after recursive synthesis, default True
        :param phase_qubit: Initial qubit to use for placing phase gates, default None
        :param **kwargs: Additional arguments passed
        :return: Constructed circuit
        """
        kwargs["full_reduce"] = True
        if architecture is None or mode == ElimMode.GAUSS_MODE:
            architecture = create_architecture(
                FULLY_CONNECTED, n_qubits=len(self.out_par[0])
            )
            mode = ElimMode.GAUSS_MODE
        else:
            mode = (
                ElimMode.GAUSS_MODE
                if mode == ElimMode.GAUSS_MODE
                else ElimMode.STEINER_MODE
            )
        n_qubits = architecture.n_qubits
        # Obtain the parities
        parities_to_reach = self.all_parities
        # Make a matrix from the parities
        matrix = Mat2(
            [
                [1 if parity[i] else 0 for parity in parities_to_reach]
                for i in range(architecture.n_qubits)
            ]
        )
        circuit = CNOT_tracker(architecture.n_qubits)
        # Make a stack - aka use the python stack >^.^<
        def recurse(
            cols_to_use, qubits_to_use, phase_qubit
        ):  # Arguments from the original paper, steiner version might only use the first
            # Check for finished columns
            cols_to_use = self._check_columns(
                matrix, circuit, cols_to_use, parities_to_reach
            )
            if cols_to_use != []:
                # Find all qubits (rows) with only 1s on the allowed parities (cols_to_use)
                qubits = [
                    i
                    for i in range(n_qubits)
                    if sum([matrix.data[i][j] for j in cols_to_use]) == len(cols_to_use)
                ]
                if len(qubits) > 1 and phase_qubit is not None:
                    # Pick the column with the most 1s to extract the steiner tree with
                    column = max(
                        cols_to_use, key=lambda c: sum([row[c] for row in matrix.data])
                    )
                    # Pick a qubit as root using the given heuristic
                    cnots = root_heuristic.to_function()(
                        architecture,
                        matrix,
                        cols_to_use,
                        qubits,
                        column,
                        phase_qubit,
                        mode=mode,
                        root_heuristic=RootHeuristic.RECURSIVE,
                        split_heuristic=SplitHeuristic.COUNT,
                    )
                    # For each returned CNOT:
                    for target, control in cnots:
                        # Place the CNOT on the circuit
                        circuit.add_gate(CNOT(control, target))
                        # Adjust the matrix accordingly - reversed elementary row operations
                        matrix.row_add(target, control)
                        # Keep track of the parities in the circuit - normal elementary row operations
                        cols_to_use = self._check_columns(
                            matrix, circuit, cols_to_use, parities_to_reach
                        )
                # After placing the cnots do recursion
                if len(cols_to_use) > 0:
                    # Choose a row to split on
                    if len(cols_to_use) == 1:
                        qubits = [
                            i
                            for i in range(n_qubits)
                            if sum([matrix.data[i][j] for j in cols_to_use]) == 1
                        ]
                    else:
                        # Ignore rows that are currently all 0s or all 1s
                        qubits = [
                            i
                            for i in range(n_qubits)
                            if sum([matrix.data[i][j] for j in cols_to_use])
                            not in [0, len(cols_to_use)]
                        ]
                    # Pick the one with the best connectivity everywhere
                    best_qubits = split_heuristic.to_function()(
                        architecture,
                        matrix,
                        cols_to_use,
                        qubits,
                        phase_qubit=phase_qubit,
                    )
                    # Pick the qubit where the recursion split will be most skewed.
                    chosen_row = max(
                        best_qubits,
                        key=lambda q: max(
                            [
                                len(
                                    [
                                        col
                                        for col in cols_to_use
                                        if matrix.data[q][col] == i
                                    ]
                                )
                                for i in [1, 0]
                            ],
                            default=-1,
                        ),
                    )
                    # Split the column into 1s and 0s in that row
                    cols1 = [
                        col for col in cols_to_use if matrix.data[chosen_row][col] == 1
                    ]
                    cols0 = [
                        col for col in cols_to_use if matrix.data[chosen_row][col] == 0
                    ]
                    rec_list = [
                        (
                            cols1,
                            qubits_to_use,
                            phase_qubit if phase_qubit is not None else chosen_row,
                        ),
                        (cols0, qubits_to_use, phase_qubit),
                    ]
                    for args in rec_list:
                        recurse(*args)

        # Put the base case into the python stack
        recurse(
            [i for i in range(len(parities_to_reach))],
            [i for i in range(n_qubits)],
            phase_qubit,
        )
        if full:
            # Calculate the final parity that needs to be added from the circuit and self.out_par
            self._obtain_final_parities(circuit, architecture, mode, **kwargs)
        # Return the circuit
        return (
            circuit,
            [i for i in range(architecture.n_qubits)],
            [i for i in range(architecture.n_qubits)],
        )

    def Ariannes_synth(
        self,
        mode,
        architecture,
        full=True,
        zeroes_rec=False,
        neighbour_path=False,
        tie_break=False,
        **kwargs,
    ) -> Tuple[Circuit, Any, Any]:
        """
        Builds CNOT circuit  to implement a set of parity transformations using recursive 
        Gaussian elimination

        :param mode: Elimination mode, controls Gaussian elimination strategy
        :param architecture: An optional of the quantum architecture to be used
        :param full: Whether to do the full parity finalization step, default True
        :param zeroes_rec: Changes the recursion heuristic to focus on zeros indead of ones, default False
        :param neighbour_path: Influences neighbor selection strategy when adding gates, default False
        :param tie_break: Enables tie-breaking heuristics when multiple choices have equal scores, default False
        :param **kwargs: Additional arguments passed
        :return: Tuple of (circuit, list_of_qubits, list_of_qubits) where the lists are the qubit indices in the architecture
        """
        kwargs["full_reduce"] = True
        if not isinstance(architecture, Architecture):
            architecture = create_architecture(
                FULLY_CONNECTED, n_qubits=len(self.out_par[0])
            )
        n_qubits = architecture.n_qubits
        # Obtain the parities
        parities_to_reach = self.all_parities
        # Make a matrix from the parities
        parity_matrix = Mat2(
            [
                [1 if parity[i] else 0 for parity in parities_to_reach]
                for i in range(architecture.n_qubits)
            ]
        )
        circuit = CNOT_tracker(architecture.n_qubits)
        cols_to_reach = self._check_columns(
            parity_matrix,
            circuit,
            list(range(len(parities_to_reach))),
            parities_to_reach,
        )
        self.prev_rows: List[int] = []

        def place_cnot(control, target):
            """
            Add CNOT gate from control qubit to target qubit and updates the parity matrix

            :param control: The control qubit
            :param target: The target qubit 
            """
            # Place the CNOT on the circuit
            circuit.add_gate(CNOT(control, target))
            # Adjust the matrix accordingly - reversed elementary row operations
            parity_matrix.row_add(target, control)

        def base_recurse(cols_to_use, qubits_to_use):
            """
            Recursively synthesize parity columns by choosing a pivot qubit and splitting columns.

            :param cols_to_use: List of indices of parity columns to synthesize
            :param qubits_to_use: List of qubit indices available for this recursion step
            """
            if qubits_to_use != [] and cols_to_use != []:
                # Select edge qubits
                vertices_to_use = [architecture.qubit2vertex(q) for q in qubits_to_use]
                vertices = architecture.non_cutting_vertices(vertices_to_use)
                qubits = [architecture.vertex2qubit(v) for v in vertices]

                # Pick the qubit where the recursion split will be most skewed.
                if zeroes_rec:
                    selection_criterium = lambda q: len(
                        [col for col in cols_to_use if parity_matrix.data[q][col] == 0]
                    )
                else:
                    selection_criterium = lambda q: max(
                        [
                            len(
                                [
                                    col
                                    for col in cols_to_use
                                    if parity_matrix.data[q][col] == i
                                ]
                            )
                            for i in [1, 0]
                        ],
                        default=-1,
                    )
                if tie_break:
                    chosen_row = self._tie_break(
                        qubits, maxelements(qubits, key=selection_criterium)
                    )
                else:
                    chosen_row = max(qubits, key=selection_criterium)
                # Split the column into 1s and 0s in that row
                cols1 = [
                    col
                    for col in cols_to_use
                    if parity_matrix.data[chosen_row][col] == 1
                ]
                cols0 = [
                    col
                    for col in cols_to_use
                    if parity_matrix.data[chosen_row][col] == 0
                ]
                base_recurse(cols0, [q for q in qubits_to_use if q != chosen_row])
                one_recurse(
                    cols1, [q for q in qubits_to_use if q != chosen_row], chosen_row
                )

        def one_recurse(cols_to_use, qubits_to_use, qubit):
            """
            Recursively synthesize parity columns where the pivot qubit row has 1s.

            :param cols_to_use: List of indices of parity columns to synthesize
            :param qubits_to_use: List of qubit indices available for this recursion step
            :param qubit: The pivot qubit index on which the current recursion is based
            """
            if cols_to_use != []:
                self._update_prev_rows([qubit])
                neighbors = [
                    q
                    for q in architecture.get_neighboring_qubits(qubit)
                    if q in qubits_to_use
                ]
                selection_criterium = lambda q: len(
                    [col for col in cols_to_use if parity_matrix.data[q][col] == 1]
                )
                indices = maxelements(neighbors, key=selection_criterium)
                if (
                    neighbour_path
                    and len(indices) > 1
                    and selection_criterium(neighbors[indices[0]]) == 0
                ):
                    # We will swap, pick the right direction to swap towards
                    # This is the neighbour with the shortest path (restricted to qubits_to_use) to a row with most ones.
                    directions = [
                        qubits_to_use[i]
                        for i in maxelements(qubits_to_use, key=selection_criterium)
                    ]  # All rows with most ones
                    path_criterium = lambda i: min(
                        directions,
                        key=lambda q: len(
                            architecture.shortest_path(neighbors[i], q, qubits_to_use)
                            or []
                        ),
                    )
                    indices = maxelements(indices, key=path_criterium, reverse=True)
                if tie_break and len(indices) > 1:
                    chosen_neighbors = self._tie_break(neighbors, indices)
                else:
                    chosen_neighbors = neighbors[indices[0]]

                # Place CNOTs if you still need to extract columns
                if (
                    sum([parity_matrix.data[chosen_neighbors][c] for c in cols_to_use])
                    != 0
                ):  # Check if adding the cnot is useful
                    place_cnot(qubit, chosen_neighbors)
                    # Might have changed the matrix.
                    cols_to_use = self._check_columns(
                        parity_matrix, circuit, cols_to_use, parities_to_reach
                    )
                else:  # Will never change the matrix
                    place_cnot(chosen_neighbors, qubit)
                    place_cnot(qubit, chosen_neighbors)
                    # Since the neighbors was all zeros, this is effectively a swap and no columns need to be checked.
                self._update_prev_rows([chosen_neighbors])
                # Split the column into 1s and 0s in that row
                cols0 = [
                    col for col in cols_to_use if parity_matrix.data[qubit][col] == 0
                ]
                cols1 = [
                    col for col in cols_to_use if parity_matrix.data[qubit][col] == 1
                ]
                base_recurse(cols0, qubits_to_use)
                one_recurse(cols1, qubits_to_use, qubit)

        base_recurse(cols_to_reach, [i for i in range(n_qubits)])

        if full:
            # Calculate the final parity that needs to be added from the circuit and self.out_par
            self._obtain_final_parities(circuit, architecture, mode, **kwargs)

        # Return the circuit
        return (
            circuit,
            [i for i in range(architecture.n_qubits)],
            [i for i in range(architecture.n_qubits)],
        )

    def _check_columns(
        self,
        parity_matrix: Mat2,
        circuit: Circuit,
        columns: List[int],
        parities_to_reach: List[Parity],
    ):
        """
        Check if any of the columns are finished (are phase gadgets over a single qubit)
        and add the corresponding phase gates to the circuit.
        """
        initial_cols = columns.copy()
        for col in initial_cols:
            column_data = [row[col] for row in parity_matrix.data]
            if sum(column_data) == 1:
                # Add phase gates where needed
                qubit = column_data.index(1)
                circuit.add_gate(ZPhase(qubit, self.zphases[parities_to_reach[col]]))
                # Remove columns from the parity_matrix if the corresponding parity was obtained
                columns.remove(col)
        return columns

    def _tie_break(self, qubits, indices):
        """
        Tie-breaking strategy to select a qubit index among multiple candidates.

        :param qubits: List of qubit indicies under consideration
        :param indices: List of indices representing tied candidates
        :return: The selected qubit index from `qubits`
        """
        for i in indices:
            if qubits[i] not in self.prev_rows:
                return qubits[i] if not qubits[i] else qubits[i]
        return qubits[indices[0]] if not qubits[indices[0]] else qubits[indices[0]]

    def _update_prev_rows(self, qubits):
        """
        Update the internal record of previously used qubits.

        :param qubits: List of qubit indicies to update into the previous rows record
        """
        # print("Updating prev rows:", self.prev_rows, qubits[0], end="\t")
        # if any(qubit in self.prev_rows for qubit in qubits):
        #    self.prev_rows = qubits
        # else:
        #    self.prev_rows += qubits
        index = max(
            [q for q in qubits if q in self.prev_rows],
            key=lambda q: self.prev_rows.index(q),
            default=-1,
        )
        if index == -1:
            self.prev_rows += qubits
        else:
            index += 2 - index % 2
            if index != len(self.prev_rows):
                self.prev_rows = self.prev_rows[index:] + qubits
            else:
                self.prev_rows = [qubits]
        # print(self.prev_rows)

    def _obtain_final_parities(
        self,
        circuit: CNOT_tracker,
        architecture: Architecture,
        mode: ElimMode,
        **kwargs,
    ):
        """
        Calculate the final parity that needs to be added from the circuit and self.out_par
        """
        current_parities = circuit.matrix
        output_parities = Mat2(
            [[1 if v else 0 for v in row] for row in self.out_par]
        )
        current_parities_inv = current_parities.inverse()
        if current_parities_inv is None:
            raise ValueError("The circuit parity is not invertible.")
        last_parities = output_parities * current_parities_inv
        # Do steiner-gauss to calculate necessary CNOTs and add those to the circuit.
        cnots = CNOT_tracker(architecture.n_qubits)
        gauss(mode, last_parities, architecture, y=cnots, **kwargs)
        for cnot in cnots.gates:
            circuit.add_gate(cnot)


def inverse_hack(matrix: Mat2) -> Optional[Tuple[Mat2, Mat2]]:
    """
    Compute the inverse of a binary matrix (over GF(2)), if matrix is not square it is extened

    :param matrix: Binary matrix to invert
    :return: Tuple of possibly extented matrix and its inverse, or None if the matrix is singular
    """
    m = matrix.copy()
    if matrix.rows() != matrix.cols():
        # Compute which columns are independent from the matrix
        rank = m.gauss(full_reduce=False)
        cols = []
        row = 0
        for col in range(m.cols()):
            if row >= m.rows() or m.data[row][col] == 0:
                cols.append(col)
            else:
                row += 1
        # Add those columns as rows to the matrix
        matrix = Mat2(
            matrix.data + [[1 if c == i else 0 for i in range(m.cols())] for c in cols]
        )
        m = matrix.copy()
    # Compute the inverse
    inv = Mat2.id(matrix.rows())
    rank = m.gauss(x=inv, full_reduce=True)
    if rank < matrix.rows():
        return None
    else:
        return matrix, inv


def partition2mat2(partition: List[Parity]) -> Mat2:
    """
    Convert a list of Parity objects into a Mat2 binary matrix.

    :param partition: List of Parity objects, each representing a binary row
    :return: Mat2 binary matrix constructed from the partition
    """
    return Mat2([parity.to_mat2_row() for parity in partition])


def mat22partition(m: Mat2) -> List[Parity]:
    """
    Convert a Mat2 binary matrix into a list of Parity objects.

    :param m: Mat2 bianry matrix
    :return: List of Parity objects representing the rows of the matrix
    """
    return [Parity(p) for p in m.data]



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import re
from fractions import Fraction
from typing import List, Dict, Tuple, Optional

from . import Circuit
from .gates import Gate, qasm_gate_table, Measurement
from ..utils import settings


class QASMParser(object):
    """Class for parsing QASM source files into circuit descriptions."""

    def __init__(self) -> None:
        self.qasm_version:int = settings.default_qasm_version
        self.gates: List[Gate] = []
        self.custom_gates: Dict[str,Circuit] = {}
        self.registers: Dict[str,Tuple[int,int]] = {}
        self.qubit_count: int = 0
        self.circuit: Optional[Circuit] = None

    def parse(self, s: str, strict:bool=True) -> Circuit:
        self.gates = []
        self.custom_gates = {}
        self.registers = {}
        self.qubit_count = 0
        self.circuit = None
        lines = s.splitlines()
        r = []
        #strip comments
        for s in lines:
            if s.find("//")!=-1:
                t = s[0:s.find("//")].strip()
            else: t = s.strip()
            if t: r.append(t)

        match = re.fullmatch(r"OPENQASM ([23])(\.\d+)?;", r[0])
        if match and match.group(1):
            self.qasm_version = int(match.group(1))
            r.pop(0)
        elif strict:
            raise TypeError("File does not start with supported OPENQASM descriptor.")

        if self.qasm_version == 2 and r[0].startswith('include "qelib1.inc";') or \
                self.qasm_version == 3 and r[0].startswith('include "stdgates.inc";'):
            r.pop(0)
        elif strict:
            raise TypeError("File is not importing standard library")

        data = "\n".join(r)
        # Strip the custom command definitions from the normal commands
        while True:
            i = data.find("gate ")
            if i == -1: break
            j = data.find("}", i)
            self.parse_custom_gate(data[i:j+1])
            data = data[:i] + data[j+1:]
        # parse the regular commands
        commands = [s.strip() for s in data.split(";") if s.strip()]
        for c in commands:
            self.gates.extend(self.parse_command(c, self.registers))

        circ = Circuit(self.qubit_count)
        circ.gates = self.gates
        self.circuit = circ
        return self.circuit

    def parse_custom_gate(self, data: str) -> None:
        data = data[5:]
        spec, body = data.split("{",1)
        if "(" in spec:
            i = spec.find("(")
            j = spec.find(")")
            if spec[i+1:j].strip():
                raise TypeError("Arguments for custom gates are currently"
                                " not supported: {}".format(data))
            spec = spec[:i] + spec[j+1:]
        spec = spec.strip()
        if " " in spec:
            name, args = spec.split(" ",1)
            name = name.strip()
            args = args.strip()
        else:
            raise TypeError("Custom gate specification doesn't have any "
                            "arguments: {}".format(data))
        registers : Dict[str,Tuple[int,int]] = {}
        qubit_count = 0
        for a in args.split(","):
            a = a.strip()
            if a in registers:
                raise TypeError("Duplicate variable name: {}".format(data))
            registers[a] = (qubit_count,1)
            qubit_count += 1

        body = body[:-1].strip()
        commands = [s.strip() for s in body.split(";") if s.strip()]
        circ = Circuit(qubit_count)
        for c in commands:
            for g in self.parse_command(c, registers):
                circ.add_gate(g)
        self.custom_gates[name] = circ

    def extract_command_parts(self, c: str) -> Tuple[str,List[Fraction],List[str]]:
        if self.qasm_version == 3:
            # Convert some OpenQASM 3 commands into OpenQASM 2 format.
            c = re.sub(r"^bit\[(\d+)] (\w+)$", r"creg \2[\1]", c)
            c = re.sub(r"^qubit\[(\d+)] (\w+)$", r"qreg \2[\1]", c)
            c = re.sub(r"^(\w+)\[(\d+)] = measure (\w+)\[(\d+)]$", r"measure \3[\4] -> \1[\2]", c)
        right_bracket = c.find(")")
        name, rest = c.split(" ", 1) if right_bracket == -1\
            else [c[:right_bracket+1], c[right_bracket+1:]]
        args = [s.strip() for s in rest.split(",") if s.strip()]
        left_bracket = name.find('(')
        phases = []
        if left_bracket != -1:
            if right_bracket == -1:
                raise TypeError(f"Mismatched bracket: {name}.")
            vals = name[left_bracket+1:right_bracket].split(',')
            phases = [self.parse_phase_arg(val) for val in vals]
            name = name[:left_bracket]
        return name, phases, args

    def parse_command(self, c: str, registers: Dict[str,Tuple[int,int]]) -> List[Gate]:
        gates: List[Gate] = []
        name, phases, args = self.extract_command_parts(c)
        if name in ("barrier","creg", "id"): return gates
        if name == "measure":
            target, result_bit = args[0].split(' -> ')
            # Extract the register name and index separately for both target and result
            _, target_idx = target.split('[')
            _, result_idx = result_bit.split('[')
            # Remove the trailing ']' and convert to int
            target_qbit = int(target_idx[:-1])
            result_register = int(result_idx[:-1])
            gate = Measurement(target_qbit, result_register)
            gates.append(gate)
            return gates
        if name in ("opaque", "if"):
            raise TypeError("Unsupported operation {}".format(c))
        if name == "qreg":
            regname, sizep = args[0].split("[",1)
            size = int(sizep[:-1])
            registers[regname] = (self.qubit_count, size)
            self.qubit_count += size
            return gates
        qubit_values = []
        is_range = False
        dim = 1
        for a in args:
            if "[" in a:
                # Split at the first '[' to handle multi-character register names
                regname, valp = a.split("[",1)
                # Remove the trailing ']' before converting to int
                val = int(valp[:-1])
                if regname not in registers: 
                    raise TypeError("Invalid register {}".format(regname))
                qubit_values.append([registers[regname][0]+val])
            else:
                if is_range:
                    if registers[a][1] != dim:
                        raise TypeError("Error in parsing {}: Register sizes do not match".format(c))
                else:
                    dim = registers[a][1]
                is_range = True
                s = registers[a][0]
                qubit_values.append(list(range(s,s + dim)))
        if is_range:
            for i in range(len(qubit_values)):
                if len(qubit_values[i]) != dim:
                    qubit_values[i] = [qubit_values[i][0]]*dim
        for j in range(dim):
            argset = [q[j] for q in qubit_values]
            if name in self.custom_gates:
                circ = self.custom_gates[name]
                if len(argset) != circ.qubits:
                    raise TypeError("Argument amount does not match gate spec: {}".format(c))
                for g in circ.gates:
                    gates.append(g.reposition(argset))
            elif name in ('x', 'y', 'z', 's', 't', 'h', 'sx'):
                if len(phases) != 0: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](argset[0])  # type: ignore # mypy can't handle Gate subclasses with different number of parameters
                gates.append(g)
            elif name in ('sdg', 'tdg', 'sxdg'):
                if len(phases) != 0: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](argset[0],adjoint=True)  # type: ignore
                gates.append(g)
            elif name in ('rx', 'ry', 'rz', 'p', 'u1'):
                if len(phases) != 1: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](argset[0],phase=phases[0])  # type: ignore
                gates.append(g)
            elif name == 'u2':
                if len(phases) != 2: raise TypeError("Invalid specification {}".format(c))
                gates.append(qasm_gate_table[name](argset[0], phases[0], phases[1]))  # type: ignore
            elif name in ('u3', 'u', 'U'):
                if len(phases) != 3: raise TypeError("Invalid specification {}".format(c))
                gates.append(qasm_gate_table[name](argset[0], phases[0], phases[1], phases[2]))  # type: ignore
            elif name in ('cx', 'CX', 'cy', 'cz', 'ch', 'csx', 'swap'):
                if len(phases) != 0: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](control=argset[0],target=argset[1])  # type: ignore
                gates.append(g)
            elif name in ('crx', 'cry', 'crz', 'cp', 'cphase', 'cu1', 'rxx', 'rzz'):
                if len(phases) != 1: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](argset[0],argset[1],phase=phases[0])  # type: ignore
                gates.append(g)
            elif name in ('ccx', 'ccz', 'cswap'):
                if len(phases) != 0: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](ctrl1=argset[0],ctrl2=argset[1],target=argset[2])  # type: ignore
                gates.append(g)
            elif name == 'cu3':
                if len(phases) != 3: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](control=argset[0],target=argset[1],theta=phases[0],phi=phases[1],rho=phases[2])  # type: ignore
                gates.append(g)
            elif name == 'cu':
                if len(phases) != 4: raise TypeError("Invalid specification {}".format(c))
                g = qasm_gate_table[name](control=argset[0],target=argset[1],theta=phases[0],phi=phases[1],rho=phases[2],gamma=phases[3])  # type: ignore
                gates.append(g)
            else:
                raise TypeError("Invalid specification: {}".format(c))
        return gates

    def parse_phase_arg(self, val):
        try:
            phase = float(val)/math.pi
        except ValueError:
            if val.find('pi') == -1: raise TypeError("Invalid specification {}".format(val))
            try:
                val = val.replace('pi', '')
                val = val.replace('*','')
                if val.find('/') != -1:
                    n, d = val.split('/',1)
                    n = n.strip()
                    if not n: n = 1
                    elif n == '-': n = -1
                    else: n = int(n)
                    d = int(d.strip())
                    phase = Fraction(n,d)
                else:
                    val = val.strip()
                    if not val: phase = 1
                    elif val == '-': phase = -1
                    else: phase = float(val)
            except: raise TypeError("Invalid specification {}".format(val))
        phase = Fraction(phase).limit_denominator(100000000)
        return phase


def qasm(s: str) -> Circuit:
    """Parses a string representing a program in QASM, and outputs a `Circuit`."""
    p = QASMParser()
    return p.parse(s, strict=False)




# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
NEW VERSION 
"""

import json
from collections import defaultdict
from fractions import Fraction

from typing import Callable, Optional, List, Dict, Tuple

from pyzx_param.utils import EdgeType, VertexType, FractionLike, toggle_edge, vertex_is_zx, toggle_vertex
from pyzx_param.graph.base import BaseGraph, VT, ET, upair
import pyzx_param.rewrite_rules.rules as rules
import pyzx_param.rewrite_rules.hrules as hrules


def match_X_spiders(
        g: BaseGraph[VT, ET],
        vertexf: Optional[Callable[[VT], bool]] = None
        ) -> List[VT]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    return [v for v in candidates if types[v] == VertexType.X]


def match_Z_spiders(
        g: BaseGraph[VT, ET],
        vertexf: Optional[Callable[[VT], bool]] = None
        ) -> List[VT]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    return [v for v in candidates if types[v] == VertexType.Z]


def color_change(g: BaseGraph[VT,ET], matches: List[VT]) -> rules.RewriteOutputType[VT,ET]:
    for v in matches:
        g.set_type(v, toggle_vertex(g.type(v)))
        for e in g.incident_edges(v):
            g.set_edge_type(e, toggle_edge(g.edge_type(e)))
    return ({}, [],[],False)


def pauli_matcher(
        g: BaseGraph[VT,ET], 
        vertexf: Optional[Callable[[VT],bool]] = None
        ) -> List[Tuple[VT,VT]]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    phases = g.phases()
    types = g.types()
    m: List[Tuple[VT,VT]] = []
    paulis = {v for v in candidates
                if phases[v] == 1 and vertex_is_zx(types[v])}
    if not paulis: return m
    while len(candidates) > 0:
        v = candidates.pop()
        if v in paulis and g.vertex_degree(v) == 2: continue
        for w in g.neighbors(v):
            if w in paulis: break
        else:
            continue
        et = g.edge_type(g.edge(v,w))
        if ((types[v] == types[w] and et == EdgeType.HADAMARD) or
            (vertex_is_zx(types[v]) and types[v] != types[w] and et == EdgeType.SIMPLE) or
            (types[v] == VertexType.H_BOX and phases[v] == 1 and (
                (et == EdgeType.SIMPLE and types[w] == VertexType.X) or
                (et == EdgeType.HADAMARD and types[w] == VertexType.Z)))
            ):
            m.append((w,v))
            candidates.difference_update(g.neighbors(v))
            candidates.difference_update(g.neighbors(w))
    return m


def pauli_push(g: BaseGraph[VT,ET],
               matches: List[Tuple[VT,VT]]
               ) -> rules.RewriteOutputType[VT,ET]:
    """Pushes a Pauli (i.e. a pi phase) through another spider."""
    rem_verts: List[VT] = []
    rem_edges: List[ET] = []
    etab: Dict[Tuple[VT,VT], List[int]] = dict()
    for w,v in matches:  # w is a Pauli and v is the spider we are gonna push it through
        if g.vertex_degree(w) == 2:
            rem_verts.append(w)
            l = list(g.neighbors(w))
            l.remove(v)
            v2 = l[0]
            et1 = g.edge_type(g.edge(v,w))
            et2 = g.edge_type(g.edge(v2,w))
            etab[upair(v,v2)] = [1,0] if et1 == et2 else [0,1]
        else:
            g.set_phase(w,0)

        new_verts = []
        if vertex_is_zx(g.type(v)):
            g.scalar.add_phase(g.phase(v))
            g.set_phase(v,(-g.phase(v)) % 2)
            t = toggle_vertex(g.type(v))
            p: FractionLike = Fraction(1)
        else:
            t = VertexType.Z
            p = 0
        for edge in g.incident_edges(v):
            st = g.edge_st(edge)
            n = st[0] if st[1] == v else st[1]
            if n == w: continue
            r = 0.5*(g.row(n) + g.row(v))
            q = 0.5*(g.qubit(n) + g.qubit(v))
            et = g.edge_type(edge)
            rem_edges.append(edge)
            w2 = g.add_vertex(t,q,r,p)
            etab[upair(v,w2)] = [1,0]
            etab[upair(n,w2)] = [1,0] if et == EdgeType.SIMPLE else [0,1]
            new_verts.append(w2)
        if not vertex_is_zx(g.type(v)): # v is H_BOX
            if len(new_verts) == 2:
                etab[upair(new_verts[0],new_verts[1])] = [0,1]
            else:
                r = (g.row(v) + sum(g.row(n) for n in new_verts)) / (len(new_verts) + 1)
                q = (g.qubit(v) + sum(g.qubit(n) for n in new_verts))/(len(new_verts)+1)
                h = g.add_vertex(VertexType.H_BOX,q,r,Fraction(1))
                for n in new_verts: etab[upair(h,n)] = [1,0]
    return (etab, rem_verts, rem_edges, False)


def match_hadamard_edge(
        g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[ET]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    return [e for e in candidates if g.edge_type(e)==EdgeType.HADAMARD]


def match_edge(
        g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[ET]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    return list(candidates)


def euler_expansion(g: BaseGraph[VT,ET], 
                    matches: List[ET]
                    ) -> rules.RewriteOutputType[VT,ET]:
    """Expands the given Hadamard-edges into pi/2 phases using its Euler decomposition."""
    types = g.types()
    phases = g.phases()
    rem_edges = []
    etab = {}
    for e in matches:
        rem_edges.append(e)
        v1,v2 = g.edge_st(e)
        if vertex_is_zx(types[v1]) and types[v1] == types[v2]:
            r = 0.5*(g.row(v1) + g.row(v2))
            q = 0.5*(g.qubit(v1) + g.qubit(v2))
            t = toggle_vertex(types[v1])
            v = g.add_vertex(t,q,r)
            etab[upair(v,v1)] = [1,0]
            etab[upair(v,v2)] = [1,0]
            if phases[v1] == Fraction(1,2) or phases[v2] == Fraction(1,2):
                g.add_to_phase(v1,Fraction(3,2))
                g.add_to_phase(v2,Fraction(3,2))
                g.set_phase(v, Fraction(3,2))
                g.scalar.add_phase(Fraction(1,4))
            else:
                g.add_to_phase(v1,Fraction(1,2))
                g.add_to_phase(v2,Fraction(1,2))
                g.set_phase(v, Fraction(1,2))
                g.scalar.add_phase(Fraction(7,4))
        else:
            r = 0.25*g.row(v1) + 0.75*g.row(v2)
            q = 0.25*g.qubit(v1) + 0.75*g.qubit(v2)
            w1 = g.add_vertex(VertexType.Z,q,r,Fraction(1,2))
            etab[upair(v2,w1)] = [1,0]
            r = 0.5*g.row(v1) + 0.5*g.row(v2)
            q = 0.5*g.qubit(v1) + 0.5*g.qubit(v2)
            w2 = g.add_vertex(VertexType.X,q,r,Fraction(1,2))
            etab[upair(w1,w2)] = [1,0]
            r = 0.75*g.row(v1) + 0.25*g.row(v2)
            q = 0.75*g.qubit(v1) + 0.25*g.qubit(v2)
            w3 = g.add_vertex(VertexType.Z,q,r,Fraction(1,2))
            etab[upair(w2,w3)] = [1,0]
            etab[upair(w3,v1)] = [1,0]
            g.scalar.add_phase(Fraction(7,4))
            
    return (etab, [], rem_edges, False)

def add_Z_identity(g: BaseGraph[VT,ET], 
        matches: List[ET]
        ) -> rules.RewriteOutputType[VT,ET]:
    rem_edges = []
    etab = {}
    for e in matches:
        rem_edges.append(e)
        et = g.edge_type(e)
        v1,v2 = g.edge_st(e)
        r = 0.5*(g.row(v1) + g.row(v2))
        q = 0.5*(g.qubit(v1) + g.qubit(v2))
        w = g.add_vertex(VertexType.Z, q,r, 0)
        etab[upair(v1,w)] = [1,0] if et == EdgeType.SIMPLE else [0,1]
        etab[upair(v2,w)] = [1,0]
    return (etab, [], rem_edges, False)

def match_bialgebra(g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[Tuple[VT,VT]]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    m = []
    types = g.types()
    phases = g.phases()
    while len(candidates) > 0:
        e = candidates.pop()
        if g.edge_type(e) != EdgeType.SIMPLE: continue
        v,w = g.edge_st(e)
        if types[v] != VertexType.X:
            v,w = w,v
        if types[v] != VertexType.X: continue
        if types[w] == VertexType.Z:
            if phases[v] != 0 or phases[w] != 0: continue
            m.append((v,w))
            for n in g.neighbors(v):
                candidates.difference_update(g.incident_edges(n))
            for n in g.neighbors(w):
                candidates.difference_update(g.incident_edges(n))
        if types[w] == VertexType.H_BOX:
            if phases[v] != 0 or phases[w] != 1: continue
            m.append((v,w))
            for n in g.neighbors(v):
                candidates.difference_update(g.incident_edges(n))
            for n in g.neighbors(w):
                candidates.difference_update(g.incident_edges(n))
    return m

def bialgebra(g: BaseGraph[VT,ET],
        matches: List[Tuple[VT,VT]]
        ) -> rules.RewriteOutputType[VT,ET]:
    rem_verts = []
    etab = {}
    for v1, v2 in matches:
        rem_verts.append(v1)
        rem_verts.append(v2)
        v = (v1,v2)
        new_verts: Tuple[List[VT],List[VT]] = ([],[]) # new vertices for v1 and v2

        for i, j in [(0, 1), (1, 0)]:
            multi_edge_found = False
            for e in g.incident_edges(v[i]):
                source, target = g.edge_st(e)
                other_vertex = source if source != v[i] else target
                if other_vertex != v[j] or multi_edge_found:
                    q = 0.4*g.qubit(other_vertex) + 0.6*g.qubit(v[i])
                    r = 0.4*g.row(other_vertex) + 0.6*g.row(v[i])
                    newv = g.add_vertex(g.type(v[j]), qubit=q, row=r)
                    g.set_phase(newv, g.phase(v[j]))
                    new_verts[i].append(newv)
                    if other_vertex == v[j]:
                        q = 0.4*g.qubit(v[i]) + 0.6*g.qubit(other_vertex)
                        r = 0.4*g.row(v[i]) + 0.6*g.row(other_vertex)
                        newv2 = g.add_vertex(g.type(v[i]), qubit=q, row=r)
                        new_verts[j].append(newv2)
                        other_vertex = newv2
                    if upair(newv, other_vertex) not in etab:
                        etab[upair(newv, other_vertex)] = [0, 0]
                    type_index = 0 if g.edge_type(e) == EdgeType.SIMPLE else 1
                    etab[upair(newv, other_vertex)][type_index] += 1
                elif i == 0: # only add new vertex once
                    multi_edge_found = True

        for n1 in new_verts[0]:
            for n2 in new_verts[1]:
                if upair(n1,n2) not in etab:
                    etab[upair(n1,n2)] = [0, 0]
                etab[upair(n1,n2)][0] += 1

        if g.type(v1) == VertexType.H_BOX or g.type(v2) == VertexType.H_BOX: # x-h bialgebra
            x_vertex = v1 if g.type(v2) == VertexType.H_BOX else v2
            g.scalar.add_power(g.vertex_degree(x_vertex)-2)
        else: # z-x bialgebra
            g.scalar.add_power((g.vertex_degree(v1)-2)*(g.vertex_degree(v2)-2))
    return (etab, rem_verts, [], False)

def match_bialgebra_op(g: BaseGraph[VT,ET],
        vertexf: Optional[Callable[[VT], bool]] = None,
        vertex_type: Optional[Tuple[VertexType, VertexType]] = None,
        edge_type: Optional[EdgeType] = None
        ) -> Optional[Tuple[List[VT], List[VT]]]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    if vertex_type is not None:
        vtype1, vtype2 = vertex_type
    else:
        vtype1, vtype2 = VertexType.Z, VertexType.X
    if edge_type is None:
        edge_type = EdgeType.SIMPLE
    type1_vertices = [v for v in candidates if g.type(v) == vtype1]
    type2_vertices = [v for v in candidates if g.type(v) == vtype2]
    if len(type1_vertices) <= 1 or len(type2_vertices) <= 1:
        return None
    #the vertices must be phase-free
    for v in type1_vertices + type2_vertices:
        if g.phase(v) != 0:
            return None
    # the vertices must have one external edge
    for v1 in type1_vertices:
        if g.vertex_degree(v1) != len(type2_vertices) + 1:
            return None
    for v2 in type2_vertices:
        if g.vertex_degree(v2) != len(type1_vertices) + 1:
            return None
    # if all type1 vertices are connected to all type2 vertices with a simple edge, then they are a match
    for v1 in type1_vertices:
        for v2 in type2_vertices:
            edges = list(g.edges(v1, v2))
            if not (len(edges) == 1 and g.edge_type(edges[0]) == edge_type):
                return None
    return type1_vertices, type2_vertices

def bialgebra_op(g: BaseGraph[VT,ET],
        matches: Tuple[List[VT], List[VT]],
        edge_type: Optional[EdgeType] = EdgeType.SIMPLE
        ) -> rules.RewriteOutputType[VT,ET]:
    def get_neighbors_and_loops(type1_vertices: List[VT], type2_vertices: List[VT]) -> Tuple[List[Tuple[VT, EdgeType]], List[EdgeType]]:
        neighbors: List[Tuple[VT, EdgeType]] = []
        loops: List[EdgeType] = []
        for v1 in type1_vertices:
            for edge in g.incident_edges(v1):
                edge_st = g.edge_st(edge)
                neighbor = edge_st[0] if edge_st[0] != v1 else edge_st[1]
                if neighbor in type2_vertices:
                    continue
                elif neighbor in type1_vertices:
                    if v1 > neighbor:
                        loops.append(g.edge_type(edge))
                else:
                    neighbors.append((neighbor, g.edge_type(edge)))
        return neighbors, loops

    def add_vertex_with_averages(vertices, g, vtype):
        average_row = sum(g.row(v) for v in vertices) / len(vertices)
        average_qubit = sum(g.qubit(v) for v in vertices) / len(vertices)
        return g.add_vertex(vtype, average_qubit, average_row)

    def update_etab(etab, new_vertex, neighbors, loops):
        for n, et in neighbors + [(new_vertex, et) for et in loops]:
            etab[upair(new_vertex, n)][0 if et == EdgeType.SIMPLE else 1] += 1

    type1_vertices, type2_vertices = matches
    neighbors1, loops1 = get_neighbors_and_loops(type1_vertices, type2_vertices)
    neighbors2, loops2 = get_neighbors_and_loops(type2_vertices, type1_vertices)

    new_vertex1 = add_vertex_with_averages(type1_vertices, g, g.type(type2_vertices[0]))
    new_vertex2 = add_vertex_with_averages(type2_vertices, g, g.type(type1_vertices[0]))

    etab: dict = defaultdict(lambda: [0, 0])
    if edge_type == EdgeType.SIMPLE:
        etab[upair(new_vertex1, new_vertex2)] = [1, 0]
    else:
        etab[upair(new_vertex1, new_vertex2)] = [0, 1]
    update_etab(etab, new_vertex1, neighbors1, loops1)
    update_etab(etab, new_vertex2, neighbors2, loops2)

    g.scalar.add_power(-(len(neighbors1)-1)*(len(neighbors2)-1))

    return (etab, type1_vertices + type2_vertices, [], False)


MATCHES_VERTICES = 1
MATCHES_EDGES = 2

operations = {
    "spider": {"text": "fuse spiders", 
               "tooltip": "Fuses connected spiders of the same color",
               "matcher": rules.match_spider_parallel, 
               "rule": rules.spider, 
               "type": MATCHES_VERTICES},
    "to_z": {"text": "change color to Z", 
               "tooltip": "Changes X spiders into Z spiders by pushing out Hadamards",
               "matcher": match_X_spiders, 
               "rule": color_change, 
               "type": MATCHES_VERTICES},
    "to_x": {"text": "change color to X", 
               "tooltip": "Changes Z spiders into X spiders by pushing out Hadamards",
               "matcher": match_Z_spiders, 
               "rule": color_change, 
               "type": MATCHES_VERTICES},
    "rem_id": {"text": "remove identity", 
               "tooltip": "Removes a 2-ary phaseless spider",
               "matcher": rules.match_ids_parallel, 
               "rule": rules.remove_ids, 
               "type": MATCHES_VERTICES},
    "id_z": {"text": "Add Z identity", 
               "tooltip": "Adds a phaseless arity 2 Z spider on the selected edges",
               "matcher": match_edge, 
               "rule": add_Z_identity, 
               "type": MATCHES_EDGES},
    "z_to_z_box": {"text": "Convert Z-spider to Z-box",
                "tooltip": "Converts a Z-spider into a Z-box",
                "matcher": rules.match_z_to_z_box_parallel,
                "rule": rules.z_to_z_box,
                "type": MATCHES_VERTICES},
    "hopf": {"text": "Hopf rule",
             "tooltip": "Applies the Hopf rule between pairs of spiders that share parallel edges",
             "matcher": rules.match_hopf,
             "rule": rules.hopf,
             "type": MATCHES_VERTICES},
    "remove_self_loops": {"text": "remove self-loops",
                        "tooltip": "Removes all self-loops on a spider",
                        "matcher": rules.match_self_loop,
                        "rule": rules.remove_self_loops,
                        "type": MATCHES_VERTICES},
    "had2edge": {"text": "Convert H-box", 
               "tooltip": "Converts an arity 2 H-box into an H-edge.",
               "matcher": hrules.match_hadamards, 
               "rule": hrules.hadamard_to_h_edge, 
               "type": MATCHES_VERTICES},
    "fuse_hbox": {"text": "Fuse H-boxes", 
               "tooltip": "Merges two adjacent H-boxes together",
               "matcher": hrules.match_connected_hboxes, 
               "rule": hrules.fuse_hboxes, 
               "type": MATCHES_EDGES},
    "mult_hbox": {"text": "Multiply H-boxes", 
               "tooltip": "Merges groups of H-boxes that have the same connectivity",
               "matcher": hrules.match_par_hbox, 
               "rule": hrules.par_hbox, 
               "type": MATCHES_VERTICES},
    "fuse_w": {"text": "fuse W nodes",
               "tooltip": "Merges two connected W nodes together",
               "matcher": rules.match_w_fusion_parallel,
                "rule": rules.w_fusion,
                "type": MATCHES_EDGES},
    "copy": {"text": "copy 0/pi spider", 
               "tooltip": "Copies a single-legged spider with a 0/pi phase through its neighbor",
               "matcher": hrules.match_copy, 
               "rule": hrules.apply_copy, 
               "type": MATCHES_VERTICES},
    "pauli": {"text": "push Pauli", 
               "tooltip": "Pushes an arity 2 pi-phase through a selected neighbor",
               "matcher": pauli_matcher, 
               "rule": pauli_push, 
               "type": MATCHES_VERTICES},
    "bialgebra": {"text": "bialgebra",
               "tooltip": "Applies the bialgebra rule to a connected pair of Z and X spiders",
               "matcher": match_bialgebra,
               "rule": bialgebra,
               "type": MATCHES_EDGES},
    "bialgebra_op": {"text": "bialgebra (inverse)",
               "tooltip": "Applies the bialgebra rule to a connected pair of Z and X spiders in the opposite direction",
               "matcher": match_bialgebra_op,
               "rule": bialgebra_op,
               "type": MATCHES_VERTICES},
    "euler": {"text": "decompose Hadamard", 
               "tooltip": "Expands a Hadamard-edge into its component spiders using its Euler decomposition",
               "matcher": match_hadamard_edge, 
               "rule": euler_expansion, 
               "type": MATCHES_EDGES},
    "lcomp": {"text": "local complementation", 
               "tooltip": "Deletes a spider with a pi/2 phase by performing a local complementation on its neighbors",
               "matcher": rules.match_lcomp_parallel, 
               "rule": rules.lcomp, 
               "type": MATCHES_VERTICES},
    "pivot": {"text": "pivot", 
               "tooltip": "Deletes a pair of spiders with 0/pi phases by performing a pivot",
               "matcher": lambda g, matchf: rules.match_pivot_parallel(g, matchf, check_edge_types=True), 
               "rule": rules.pivot, 
               "type": MATCHES_EDGES}
}


def operations_to_js() -> str:
    global operations
    return json.dumps({k:
            {"active":False, 
             "text":v["text"], 
             "tooltip":v["tooltip"]
            } for k,v in operations.items()})



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2021 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from tqdm import tqdm
import random
import math
import numpy as np

from .congruences import uniform_weights, apply_rand_lc, apply_rand_pivot
from .scores import g_wgc
import sys
if __name__ == '__main__':
    sys.path.append('..')
from pyzx_param.simplify import full_reduce


"""
This module contains an implementation of simulated annealing over ZX-diagrams. Equivalent ZX-diagrams are generated using the congruences defined in congruences.py. The default energy function is defined in scores.py. The default goal of this approach is to reduce the 2-qubit count of a fully-simplified ZX-diagram (i.e., of that circuit obtained via extraction).
"""


__all__ = ['anneal']

# simulated annealing
def anneal(g, iters=1000,
           temp=25,
           cool=0.005,
           score=g_wgc,
           cong_ps=[0.5, 0.5],
           lc_select=uniform_weights,
           pivot_select=uniform_weights,
           full_reduce_prob=0.1,
           reset_prob=0.0,
           quiet=False
):
    """
    Performs simulated annealing over ZX-diagram to minimize energy function.

    Equivalent ZX-diagrams are generated using congruences from `congruences.py`. Default goal is to reduce the 2-qubit count of a fully-simplified ZX-diagram.

    :param g: Initial ZX-diagram to optimize.
    :param iters: Number of iterations to perform (default is 1000).
    :param temp: Initial temperature for annealing (default is 25).
    :param cool: Cooling rate for temperature reduction (default is 0.005).
    :param score: Function to evaluate the energy of a diagram (default is `g_wgc`).
    :param cong_ps: Probabilities for selecting congruence methods (default is [0.5, 0.5]).
    :param lc_select: Function to select vertices for local complementation (default is `uniform_weights`).
    :param pivot_select: Function to select edges for pivoting (default is `uniform_weights`).
    :param full_reduce_prob: Probability of applying full reduction (default is 0.1).
    :param reset_prob: Probability of resetting to the best state (default is 0.0).
    :param quiet: If True, suppresses progress output (default is False).
    :return: A tuple containing the best ZX-diagram found and a list of best scores over iterations.
    """

    g_best = g.copy()
    sz = score(g_best)
    sz_best = sz

    best_scores = list()

    for i in tqdm(range(iters), desc="annealing...", disable=quiet):

        g1 = g.copy()

        # cong_method = np.random.choice(["LC", "PIVOT"], 1, p=cong_ps)[0]
        cong_method = "PIVOT"

        if cong_method == "PIVOT":
            apply_rand_pivot(g1, weight_func=pivot_select)
        else:
            apply_rand_lc(g1, weight_func=lc_select)

        # probabilistically full_reduce:
        if random.uniform(0, 1) < full_reduce_prob:
            full_reduce(g1)
        sz1 = score(g1)

        best_scores.append(sz_best)

        if temp != 0: temp *= 1.0 - cool

        if sz1 < sz or \
            (temp != 0 and random.random() < math.exp((sz - sz1)/temp)):

            sz = sz1
            g = g1.copy()
            if sz < sz_best:
                g_best = g.copy()
                sz_best = sz
        elif random.uniform(0, 1) < reset_prob:
            g = g_best.copy()

    return g_best, best_scores



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import numpy as np
from galois import GF2
from typing import Union, Any, Tuple, List, Optional, Dict, cast
from typing_extensions import Literal

from .circuit.gates import CNOT

try:
    from .linalg_c import do_gauss as gauss_fast # type: ignore
except ImportError:
   gauss_fast = None

Z2 = Literal[0,1]
MatLike = List[List[Z2]]

class Mat2(object):
    """A matrix over Z2, with methods for multiplication, primitive row and column
    operations, Gaussian elimination, rank, and epi-mono factorisation."""
    
    @staticmethod
    def id(n: int) -> 'Mat2':
        return Mat2([[1 if i == j else 0
            for j in range(n)] 
              for i in range(n)])
    @staticmethod
    def zeros(m:int, n: int) -> 'Mat2':
        return Mat2([[0
            for j in range(n)] 
              for i in range(m)])
    @staticmethod
    def unit_vector(d: int, i: int) -> 'Mat2':
        return Mat2([[1 if j == i else 0] for j in range(d)])

    def __init__(self, data: MatLike):
        self.data: MatLike = data
    def __mul__(self, m: 'Mat2') -> 'Mat2':
        return Mat2([[cast(Z2, sum(self.data[i][k] * m.data[k][j] for k in range(len(m.data))) % 2)
                      for j in range(len(m.data[0]))] for i in range(len(self.data))])
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Mat2): return False
        if self.rows() != other.rows() or self.cols() != other.cols(): return False
        return all(self.data[i][j] == other.data[i][j] for i in range(len(self.data)) for j in range(len(self.data[i])))
    def __str__(self) -> str:
        return "\n".join("[ " + 
            "  ".join(str(value) for value in row) +
            " ]" for row in self.data)
    def __repr__(self) -> str:
        return str(self)
    def __getitem__(self, key: Tuple[Union[int,slice],Union[int,slice]]) -> Union['Mat2',Z2]:
        # For a pair of indices: if either is a slice, return the
        # selected sub-matrix. Otherwise, return the selected element.
        if isinstance(key,tuple):
            rs,cs = key
            if isinstance(rs,slice) or isinstance(cs,slice):
                if not isinstance(rs,slice): rs = slice(rs,rs+1)
                if not isinstance(cs,slice): cs = slice(cs,cs+1)
                return Mat2([row[cs] for row in self.data[rs]])
            else:
                return self.data[rs][cs]
        else:
            raise IndexError("Expected a pair of indices/slices.")
    def __setitem__(self, key, val):
        # For a pair of indices: if either is a slice, expect a Mat2
        # and overwrite the selected sub-matrix. Otherwise, expect
        # Z2 and overwrite the selected element.
        if isinstance(key,tuple):
            rs,cs = key

            if isinstance(val,Mat2):
                d = val.data
            else:
                d = [[val]]

            if isinstance(rs,slice):
                rr = range(*rs.indices(self.rows()))
            else:
                rr = range(rs,rs+1)

            if isinstance(cs,slice):
                cr = range(*cs.indices(self.cols()))
            else:
                cr = range(cs,cs+1)

            for i,iin in enumerate(rr):
                for j,jin in enumerate(cr):
                    self.data[iin][jin] = d[i][j]
        else:
            raise IndexError("Expected a pair of indices/slices.")
            

    def copy(self) -> 'Mat2':
        return Mat2([list(row) for row in self.data])
    def transpose(self) -> 'Mat2':
        return Mat2([[self.data[i][j] for i in range(self.rows())] for j in range(self.cols())])
    def rows(self) -> int:
        return len(self.data)
    def cols(self) -> int:
        return len(self.data[0]) if (len(self.data) != 0) else 0
    def row_add(self, r0: int, r1: int) -> None:
        """Add r0 to r1"""
        row1 = self.data[r0]
        row2 = self.data[r1]
        for i, v in enumerate(row1):
            if v:
                row2[i] = 0 if row2[i] else 1
    def col_add(self, c0: int, c1: int) -> None:
        """Add r0 to r1"""
        for i in range(self.rows()):
            d = self.data[i]
            if d[c0]:
                d[c1] = 0 if d[c1] else 1
    def row_swap(self, r0: int, r1: int) -> None:
        """Swap the rows r0 and r1"""
        r = self.data[r0]
        self.data[r0] = self.data[r1]
        self.data[r1] = r
    def col_swap(self, c0: int, c1: int) -> None:
        """Swap the columns c0 and c1"""
        for r in range(self.rows()):
            v = self.data[r][c0]
            self.data[r][c0] = self.data[r][c1]
            self.data[r][c1] = v
    
    def permute_rows(self, p: List[int]) -> None:
        """Permute the rows of the matrix according to the permutation p."""
        self.data = [self.data[i] for i in p]
    def permute_cols(self, p: List[int]) -> None:
        """Permute the columns of the matrix according to the permutation p."""
        self.data = [[self.data[i][j] for j in p] for i in range(self.rows())]
    
    def gauss(self, full_reduce:bool=False, x:Any=None, y:Any=None, blocksize:int=6, pivot_cols:List[int]=[]) -> int:
        """Compute the echelon form. Returns the number of non-zero rows in the result, i.e.
        the rank of the matrix.

        The parameter 'full_reduce' determines whether to compute the full row-reduced form,
        useful e.g. for matrix inversion and CNOT circuit synthesis.

        The parameter 'blocksize' gives the size of the blocks in a block matrix for
        performing Patel/Markov/Hayes optimization, see:

        K. Patel, I. Markov, J. Hayes. Optimal Synthesis of Linear Reversible
        Circuits. QIC 2008

        If blocksize is given as self.cols(), then
        this is equivalent to just eliminating duplicate rows before doing normal
        Gaussian elimination.

        Contains two convenience parameters for saving the primitive row operations. Suppose
        the row-reduced form of m is computed as:

        g * m = m'

        Then, x --> g * x and y --> y * g^-1.

        Note x and y need not be matrices. x can be any object that implements the method
        row_add(), and y any object that implements col_add().
        """

        rows = self.rows()
        cols = self.cols()
        #pivot_cols = []
        pivot_row = 0
        for sec in range(math.ceil(cols / blocksize)):
            i0 = sec * blocksize
            i1 = min(cols, (sec+1) * blocksize)

            # search for duplicate chunks of 'blocksize' bits and eliminate them
            chunks: Dict[Tuple[Z2,...],int] = dict()
            for r in range(pivot_row, rows):
                t = tuple(self.data[r][i0:i1])
                if not any(t): continue
                if t in chunks:
                    #print('hit (down)', r, chunks[t], t, i0, i1)
                    self.row_add(chunks[t], r)
                    if x is not None: x.row_add(chunks[t], r)
                    if y is not None: y.col_add(r, chunks[t])
                else:
                    chunks[t] = r

            p = i0
            while p < i1:
                for r0 in range(pivot_row, rows):
                    if self.data[r0][p] != 0:
                        if r0 != pivot_row:
                            self.row_add(r0, pivot_row)
                            if x is not None: x.row_add(r0, pivot_row)
                            if y is not None: y.col_add(pivot_row, r0)

                        for r1 in range(pivot_row+1, rows):
                            # TODO: remove pivot_row != r1 and test
                            if pivot_row != r1 and self.data[r1][p] != 0:
                                self.row_add(pivot_row, r1)
                                if x is not None: x.row_add(pivot_row, r1)
                                if y is not None: y.col_add(r1, pivot_row)
                        #if full_reduce:
                        pivot_cols.append(p)
                        pivot_row += 1
                        break
                p += 1
        
        rank = pivot_row

        if full_reduce:
            pivot_row -= 1
            pivot_cols1 = pivot_cols.copy()

            for sec in range(math.ceil(cols / blocksize) - 1, -1, -1):
                i0 = sec * blocksize
                i1 = min(cols, (sec+1) * blocksize)

                # search for duplicate chunks of 'blocksize' bits and eliminate them
                chunks = dict()
                for r in range(pivot_row, -1, -1):
                    t = tuple(self.data[r][i0:i1])
                    if not any(t): continue
                    if t in chunks:
                        #print('hit (up)', r, chunks[t], t, i0, i1)
                        self.row_add(chunks[t], r)
                        if x is not None: x.row_add(chunks[t], r)
                        if y is not None: y.col_add(r, chunks[t])
                    else:
                        chunks[t] = r

                while len(pivot_cols1) != 0 and i0 <= pivot_cols1[-1] < i1:
                    pcol = pivot_cols1.pop()
                    for r in range(0, pivot_row):
                        if self.data[r][pcol] != 0:
                            self.row_add(pivot_row, r)
                            if x is not None: x.row_add(pivot_row, r)
                            if y is not None: y.col_add(r, pivot_row)
                    pivot_row -= 1

        return rank

    def rank(self) -> int:
        """Returns the rank of the matrix."""
        m = self.copy()
        return m.gauss()

    def factor(self) -> Tuple['Mat2','Mat2']:
        """Produce a factorisation m = m0 * m1, where

        m0.cols() = m1.rows() = m.rank()
        """
        
        # identity matrix
        m0 = Mat2.id(self.rows())
        
        # copy of m (aka self)
        m1 = self.copy()
        
        # produce m1 := g * m and m0 := g^-1. Hence, m0 * m1 = m.
        rank = m1.gauss(y = m0)
        
        # throw away zero rows in m1, and their corresponding columns in m0
        m0 = Mat2([[row[i] for i in range(rank)] for row in m0.data])
        m1 = Mat2([m1.data[i] for i in range(rank)])
        return (m0, m1)

    def inverse(self) -> Optional['Mat2']:
        """Returns the inverse of m is invertible and None otherwise."""
        if self.rows() != self.cols(): return None
        m = self.copy()
        inv = Mat2.id(self.rows())
        rank = m.gauss(x=inv, full_reduce=True)
        if rank < self.rows(): return None
        else: return inv

    def solve(self, b: 'Mat2') -> Optional['Mat2']:
        """Return a vector x such that M * x = b, or None if there is no solution."""
        m = self.copy()
        b1 = b.copy()
        rank = m.gauss(x=b1, full_reduce=True)

        # check for inconsistencies and set x to a
        #  particular solution
        x = Mat2.zeros(m.cols(),1)
        for i,row in enumerate(m.data):
            got_pivot = False
            for j,v in enumerate(row):
                if v != 0:
                    got_pivot = True
                    x.data[j][0] = b1.data[i][0]
                    break
            # zero LHS with non-zero RHS = no solutions
            if not got_pivot and b1.data[i][0] != 0:
                return None
        return x

        # i = b1.rows() - 1
        # while i > rank - 1:
        #     if b1.data[i][0] != 0:
        #         return None
        #     i -= 1
        # if x.rows() > m.cols():
        #     x.data = x.data[:m.cols()]
        # else:
        #     x.data = x.data + [[0]]*(m.cols()-x.rows())
        # return x

    def nullspace(self, should_copy:bool=True) -> List[List[Z2]]:
        """Returns a list of non-zero vectors that span the nullspace
        of the matrix. If the matrix has trivial kernel it returns the empty list."""
        if gauss_fast:
            data = gauss_fast(self.data,1)
            m = Mat2(data)
        elif should_copy:
            m = self.copy()
            m.gauss(full_reduce=True)
        else:
            m = self
            m.gauss(full_reduce=True)
        cols = self.cols()
        nonpivots = list(range(cols))
        pivots = []
        for i, r in enumerate(m.data):
            for j in range(cols):
                if r[j]:
                    nonpivots.remove(j)
                    pivots.append(j)
                    break
        vectors:List[List[Z2]] = []
        for n in nonpivots:
            v:List[Z2] = [0]*cols
            v[n] = 1
            for r, p in zip(m.data, pivots):
                if r[n]: v[p] = 1
            vectors.append(v)
        return vectors

    def to_cnots(self, optimize: bool = False, use_log_blocksize: bool = False) -> List[CNOT]:
        """Returns a list of CNOTs that implements the matrix as a reversible circuit of qubits."""
        cn: Optional[CNOTMaker]
        if not optimize:
            cn = CNOTMaker()
            blocksize = 5
            if use_log_blocksize:
                blocksize = int(math.log2(self.rows()))
            self.copy().gauss(full_reduce=True,x=cn, blocksize=blocksize)
        else:
            best = 1000000
            best_cn = None
            for size in range(1,self.rows() + 1):
                cn = CNOTMaker()
                assert cn is not None
                self.copy().gauss(full_reduce=True,x=cn, blocksize=size)
                if len(cn.cnots) < best:
                    best = len(cn.cnots)
                    best_cn = cn
            cn = best_cn
        assert cn is not None
        return cn.cnots # list(reversed(cn.cnots)) 


class CNOTMaker(object):
    def __init__(self) -> None:
        self.cnots: List[CNOT] = []
    def row_add(self, r1:int, r2:int) -> None:
        self.cnots.append(CNOT(r2,r1))


BASE = 64


class REF:
    """
    A class to efficiently compute cut-ranks for multiple partitions.
    It stores the row echelon form of the adjacency matrix, allowing fast row addition and column deletion.
    Primitive row operations are optimised using int64 arithmetic.
    Initially, all vertices are in the right part.
    The function take(v) moves a vertex to the left, and rank() returns the cut-rank.
    """

    def __init__(self, graph):
        self.n = len(graph)
        self.N = (self.n + BASE - 1) // BASE * BASE
        self.taken = [False] * self.N
        self.graph = np.hstack([graph, np.zeros((self.n, self.N - self.n), dtype=graph.dtype)])
        self.graph = np.packbits(self.graph, axis=-1, bitorder='little').view(np.uint64)
        self.mat = []
        self.pivot_cols = []

    def _add_row(self, row_arr, start_col=0):
        pivot_row = 0
        for col in range(start_col, self.N):
            if self.taken[col]:
                continue
            while pivot_row < len(self.pivot_cols) and self.pivot_cols[pivot_row] < col:
                pivot_row += 1
            col_loc, col_rem = divmod(col, BASE)
            if (row_arr[col_loc] >> np.uint64(col_rem)) & np.uint64(1):
                if pivot_row < len(self.pivot_cols) and self.pivot_cols[pivot_row] == col:
                    row_arr ^= self.mat[pivot_row]
                else:
                    self.mat.insert(pivot_row, row_arr)
                    self.pivot_cols.insert(pivot_row, col)
                    break

    def take(self, col):
        """
        Move a vertex from the right part to the left in O(n^2) time.
        :param col: vertex id
        """
        self.taken[col] = True
        if col in self.pivot_cols:
            pivot_row = self.pivot_cols.index(col)
            self.pivot_cols.pop(pivot_row)
            row_arr = self.mat[pivot_row]
            self.mat.pop(pivot_row)
            self._add_row(row_arr, col + 1)
        self._add_row(self.graph[col])

    def rank(self):
        """
        Cut-rank of the partition, computed in O(1) time.
        """
        return len(self.pivot_cols)


def rank_factorize(A: np.ndarray) -> Tuple[int, np.ndarray, np.ndarray]:
    """
    Rank factorisation of a matrix over GF(2).
    :param A: binary matrix of shape (n, m)
    :return r: int - rk(A)
            L: binary matrix of shape (n, n) - invertible matrix whose first r columns represent the first factor
            R: binary matrix of shape (m, m) - invertible matrix whose first r rows represent the second factor
    """
    assert len(A.shape) == 2 and A.dtype == np.int8
    n, m = A.shape
    A1 = A.copy()
    L, R = np.eye(n, dtype=np.int8), np.eye(m, dtype=np.int8)
    pivots = []
    r = 0
    for j in range(m):
        if A1[r:, j].any():
            pivots.append(j)
            i = r + np.where(A1[r:, j])[0][0]
            A1[[r, i]] = A1[[i, r]]
            L[:, [r, i]] = L[:, [i, r]]
            for k in range(r + 1, n):
                if A1[k, j]:
                    A1[k] ^= A1[r]
                    L[:, r] ^= L[:, k]
            r += 1
    for i, j in list(enumerate(pivots))[::-1]:
        for k in range(i - 1, -1, -1):
            if A1[k, j]:
                A1[k] ^= A1[i]
                L[:, i] ^= L[:, k]
    for i, j in enumerate(pivots):
        for k in range(j + 1, m):
            if A1[i, k]:
                A1[:, k] ^= A1[:, j]
                R[j] ^= R[k]
    for i, j in enumerate(pivots):
        A1[:, [i, j]] = A1[:, [j, i]]
        R[[i, j]] = R[[j, i]]
    return A1.sum(), L, R


def generalized_inverse(A: np.ndarray) -> np.ndarray:
    """
    Compute the generalised inverse of a matrix over GF(2).
    :param A: binary matrix of shape (n, m)
    :return A^g: binary matrix of shape (m, n) - a matrix satisfying A * A^g * A = A
    """
    r, U, V = rank_factorize(A)
    return (np.linalg.inv(GF2(V))[:, :r] @ np.linalg.inv(GF2(U))[:r]) == 1
